# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The kernels sub-package."""

from .alias import matern_five_halves as matern_five_halves
from .alias import matern_one_half as matern_one_half
from .alias import matern_three_halves as matern_three_halves
from .alias import periodic as periodic
from .alias import rbf as rbf
from .base import Kernel as Kernel
from .base import from_kernel_function as from_kernel_function
from .transformed import combine as combine
from .transformed import scale as scale
